Imports System.IO
Imports System.Security.Cryptography

Module CryptoStuff
    Private m_RngRand As New RNGCryptoServiceProvider()

    ' Encrypt a string into a hex byte string.
    Public Function EncryptToString(ByVal plain_text As String, ByVal password As String, ByVal salt As String) As String
        Return BytesToHexString(Encrypt(plain_text, password, salt))
    End Function

    ' Decrypt a string from a hex byte string.
    Public Function DecryptFromString(ByVal encrypted_bytes_string As String, ByVal password As String, ByVal salt As String) As String
        Return Decrypt(HexStringToBytes(encrypted_bytes_string), password, salt)
    End Function

    ' Encrypt a string into an array of bytes.
    Private Function Encrypt(ByVal plain_text As String, ByVal password As String, ByVal salt As String) As Byte()
        Dim ascii_encoder As New System.Text.ASCIIEncoding()
        Dim plain_bytes As Byte() = ascii_encoder.GetBytes(plain_text)
        Return CryptBytes(password, plain_bytes, salt, True)
    End Function

    ' Decrypt a string from an array of bytes.
    Private Function Decrypt(ByVal encrypted_bytes() As Byte, ByVal password As String, ByVal salt As String) As String
        Dim decrypted_bytes() As Byte = CryptBytes(password, encrypted_bytes, salt, False)
        Dim ascii_encoder As New System.Text.ASCIIEncoding()
        Return ascii_encoder.GetChars(decrypted_bytes)
    End Function

    ' Convert bytes into a string.
    Private Function BytesToHexString(ByVal bytes() As Byte) As String
        Dim result As String = ""
        For Each b As Byte In bytes
            result &= " " & Hex(b).PadLeft(2, "0")
        Next b
        If result.Length > 0 Then result = result.Substring(1)
        Return result
    End Function

    ' Convert a hex string into bytes.
    Private Function HexStringToBytes(ByVal str As String) As Byte()
        str = str.Replace(" ", "")
        Dim max_byte As Integer = str.Length \ 2 - 1
        Dim bytes(max_byte) As Byte
        For i As Integer = 0 To max_byte
            bytes(i) = CByte("&H" & str.Substring(2 * i, 2))
        Next i
        Return bytes
    End Function

    ' Use the password to generate key bytes.
    Private Sub MakeKeyAndIV(ByVal password As String, ByVal salt() As Byte, ByVal key_size_bits As Integer, ByVal block_size_bits As Integer, ByRef key() As Byte, ByRef iv() As Byte)
        Dim derive_bytes As New Rfc2898DeriveBytes(password, salt, 1337)
        key = derive_bytes.GetBytes(key_size_bits \ 8)
        iv = derive_bytes.GetBytes(block_size_bits \ 8)
    End Sub

    ' Encrypt or decrypt a byte array.
    Private Function CryptBytes(ByVal password As String, ByVal in_bytes() As Byte, ByVal salt As String, Optional ByVal encrypt As Boolean = True) As Byte()
        ' Make a triple DES service provider.
        Dim des_provider As New TripleDESCryptoServiceProvider

        ' Find a valid key size for this provider.
        Dim key_size_bits As Integer = 0
        For i As Integer = 1024 To 1 Step -1
            If des_provider.ValidKeySize(i) Then
                key_size_bits = i
                Exit For
            End If
        Next i
        Debug.Assert(key_size_bits > 0)

        ' Get the block size for this provider.
        Dim block_size_bits As Integer = des_provider.BlockSize

        ' Convert the salt into bytes.
        Dim salt_bytes As Byte() = HexStringToBytes(salt)

        ' Generate the key and initialization vector.
        Dim key As Byte() = Nothing
        Dim iv As Byte() = Nothing
        MakeKeyAndIV(password, salt_bytes, key_size_bits, block_size_bits, key, iv)

        ' Make the encryptor or decryptor.
        Dim crypto_transform As ICryptoTransform
        If encrypt Then
            crypto_transform = des_provider.CreateEncryptor(key, iv)
        Else
            crypto_transform = des_provider.CreateDecryptor(key, iv)
        End If

        ' Create the output stream.
        Dim out_stream As New MemoryStream()

        ' Attach a crypto stream to the output stream.
        Dim crypto_stream As New CryptoStream( _
           out_stream, crypto_transform, _
           CryptoStreamMode.Write)

        ' Write the bytes into the CryptoStream.
        crypto_stream.Write(in_bytes, 0, in_bytes.Length)
        Try
            crypto_stream.FlushFinalBlock()
        Catch crypto_ex As CryptographicException
            ' Ignore this one. The password is bad.
        Catch ex As Exception
            ' Re-raise this one.
            Throw ex
        End Try

        ' Save the result.
        Dim result() As Byte = out_stream.ToArray()

        ' Close the stream.
        Try
            crypto_stream.Close()
        Catch crypto_ex As CryptographicException
            ' Ignore this one. The password is bad.
        Catch ex As Exception
            ' Re-raise this one.
            Throw ex
        End Try
        out_stream.Close()

        Return result
    End Function

    ' Return a pseudo-random salt.
    Public Function RandomSalt() As String
        ' Pick a random number of bytes.
        Dim num_bytes As Integer = RandomConstrainedInteger(10, 20)

        ' Fill the salt array.
        Dim salt_bytes(num_bytes - 1) As Byte
        m_RngRand.GetBytes(salt_bytes)

        ' Return the salt converted into a hex string.
        Return BytesToHexString(salt_bytes)
    End Function

    ' Return a random integer between a min and max value.
    Public Function RandomConstrainedInteger(ByVal min As Integer, ByVal max As Integer) As Integer
        Dim num_values As Integer = max - min + 1
        If num_values < 1 Then Throw New ArgumentException("Cannot generate a number between " & max & " & and " & min)
        If num_values > 256 Then Throw New ArgumentException("The values " & max & " & and " & min & " are too far apart for function RandomConstrainedInteger")

        ' Get random numbers until one falls within the allowed range.
        ' This preserves the random distribution of values.
        Dim one_byte() As Byte = {0}
        Do
            m_RngRand.GetBytes(one_byte)
        Loop Until one_byte(0) < num_values

        Return min + one_byte(0)
    End Function
End Module
